(**********************************************************************
 * ARBRES AVL
 * par Stphane GLONDU (Avril 2004)
 **********************************************************************)

(** Dfinition des types **********************************************)

type clef = int
let max_diff = try int_of_string (Sys.getenv "AVL_MAX_DIFF") with _ -> 1

type 'a t =
  | Vide
  | Noeud of ('a t * int * clef * 'a * 'a t)

let hauteur = function
  | Vide -> -1
  | Noeud (_,h,_,_,_) -> h

(** Fonctions pour modules externes ***********************************)

let vide = Vide
and gauche = function
  | Noeud (g,_,_,_,_) -> g
  | _ -> raise Not_found
and droite = function
  | Noeud (_,_,_,_,d) -> d
  | _ -> raise Not_found
and clef = function
  | Noeud (_,_,x,_,_) -> x
  | _ -> raise Not_found
and dcolor c _ = c.(0)

(** Vrification et recherche *****************************************)

let verifier arbre =
  let rec aux inf sup = function
    | Vide -> (0, -1)
    | Noeud (g,h,x,_,d) ->
        if inf < x && x < sup && begin
          let hg = hauteur g and hd = hauteur d in
          if hg > hd then hg - hd <= max_diff && h = hg+1
          else hd-hg <= max_diff && h = hd+1
        end then begin
          let ng = fst (aux inf x g) in
	  let nd = fst (aux x sup d) in
          (1+ng+nd, h)
        end else invalid_arg "arbre AVL incorrect"
  in aux min_int max_int arbre

let rec rechercher x = function
  | Vide -> raise Not_found
  | Noeud (g,_,r,_,_) when x < r -> rechercher x g
  | Noeud (_,_,r,_,d) when r < x -> rechercher x d
  | Noeud (_,_,_,v,_) -> v

(** Rquilibrage *******************************************************)

let construire g x v d =
  let h = 1 + max (hauteur g) (hauteur d) in
  Noeud (g,h,x,v,d)

let equilibrer g x v d =
  let hg = hauteur g and hd = hauteur d in
  let error () = failwith "equilibrer" in
  if hg > hd + max_diff then begin
    match g with
    | Vide -> error ()
    | Noeud (gg,_,gx,gv,gd) ->
       	if hauteur gg >= hauteur gd then
          (* Cas 1 *)
          construire gg gx gv (construire gd x v d)
        else begin
          (* Cas 2 *)
          match gd with
          | Vide -> error ()
          | Noeud (gdg,_,gdx,gdv,gdd) ->
              construire (construire gg gx gv gdg) gdx gdv (construire gdd x v d)
        end
  end else if hd > hg + max_diff then begin
    match d with
    | Vide -> error ()
    | Noeud (dg,_,dx,dv,dd) ->
	if hauteur dd >= hauteur dg then
	  (* Cas 1' *)
	  construire (construire g x v dg) dx dv dd
	else begin
	  (* Cas 2' *)
	  match dg with
	  | Vide -> error ()
	  | Noeud (dgg,_,dgx,dgv,dgd) ->
	      construire (construire g x v dgg) dgx dgv (construire dgd dx dv dd)
	end
  end else construire g x v d

(** Insertion *********************************************************)

let rec inserer nx nv = function
  | Vide -> Noeud (Vide, 0, nx, nv, Vide)
  | Noeud (g,_,x,v,d) when nx < x -> equilibrer (inserer nx nv g) x v d
  | Noeud (g,_,x,v,d) when x < nx -> equilibrer g x v (inserer nx nv d)
  | Noeud (g,h,_,_,d) -> Noeud (g,h,nx,nv,d)

(** Suppression *******************************************************)

let rec minimum = function
  | Vide -> raise Not_found
  | Noeud (Vide, _, x, d, _) -> (x, d)
  | Noeud (g,_,_,_,_) -> minimum g

let rec supprimer_minimum = function
  | Vide -> failwith "supprimer_minimum"
  | Noeud (Vide, _, _, _, d) -> d
  | Noeud (g,_,x,v,d) -> equilibrer (supprimer_minimum g) x v d

let fusionner g d =
  match (g, d) with
  | (Vide, a) | (a, Vide) -> a
  | _ ->
      let (x,v) = minimum d in
      equilibrer g x v (supprimer_minimum d)

let rec supprimer nx = function
  | Vide -> Vide
  | Noeud (g,_,x,v,d) when nx < x -> equilibrer (supprimer nx g) x v d
  | Noeud (g,_,x,v,d) when x < nx -> equilibrer g x v (supprimer nx d)
  | Noeud (g,_,_,_,d) -> fusionner g d

(**********************************************************************)
