(** Dessin d'arbres
    @author Stphane Glondu
    @since 2004/04/05
*)

open Graphics

module type TArbres = sig
  type 'a t
  val gauche : 'a t -> 'a t
  val droite : 'a t -> 'a t
  val clef : 'a t -> int
  val dcolor : 'a array -> 'b t -> 'a
end

let bw_mode = try Sys.getenv "BW_MODE" <> "" with _ -> false

(****************************************************************************************************)

module type TDessineur = sig
  type 'a t
  val dessiner : int -> int -> int -> int -> 'a t -> unit
  val dessiner_std : 'a t -> unit
  val calculer_dim : 'a t -> int * int
end

module MakeDessineur (Arbres : TArbres) = struct

  open Graphics
  type 'a t = 'a Arbres.t
  type 'a dxt = Vide | Noeud of ('a dxt * 'a t * (int * int) list * 'a dxt)

  let grey = rgb 220 220 220
  let couleurs = if bw_mode then
    [| (black, grey); (black, white); (white, black) |]
  else
    [| (black, grey); (white,   red); (white, black) |]

  let rec calculer_dx arbre =
    try
      match (calculer_dx (Arbres.gauche arbre), calculer_dx (Arbres.droite arbre)) with
      | Vide, Vide -> Noeud (Vide, arbre, [], Vide)
      | Vide, (Noeud (_,_,dv,_) as d) ->
	  let nouv = List.map (fun (x,y) -> (x+1,y+1)) dv in
	  Noeud (Vide, arbre, (1,1)::nouv, d)
      | (Noeud (_,_,gv,_) as g), Vide ->
	  let nouv = List.map (fun (x,y) -> (x-1,y-1)) gv in
	  Noeud (g, arbre, (-1,-1)::nouv, Vide)
      | (Noeud (_,_,gv,_) as g), (Noeud (_,_,dv,_) as d) ->
	  let rec sparation dj = function
	    | ([], _) | (_, []) -> dj
	    | ((_,x)::q, (y,_)::r) -> sparation (max dj ((x-y+3)/2)) (q,r)
	  in
	  let dx = sparation 1 (gv, dv) in
	  let rec recombine = function
	    | ([], []) -> []
	    | ([], r) -> List.map (fun (x,y) -> (x+dx,y+dx)) r
	    | (r, [] ) -> List.map (fun (x,y) -> (x-dx,y-dx)) r
	    | ((x1,y1)::q, (x2,y2)::r) -> (x1-dx,y2+dx) :: (recombine (q,r))
	  in
	  Noeud (g, arbre, (-dx,dx) :: (recombine (gv, dv)), d)
    with Not_found -> Vide

  let calculer_dim arbre =
    match calculer_dx arbre with
    | Vide -> (0, 0)
    | Noeud (_,_,v,_) as t ->
	let (minim, maxim) = List.fold_left (fun (a,b) (c,d) -> (min a c, max b d)) (0,0) v in
	let largeur = maxim-minim+1 in
	let hauteur = List.length v in
	(largeur, hauteur)

  let dessiner dy rad x y arbre =
    let rec aux x y = function
      | Vide -> ()
      | Noeud (g, v, dxlist, d) ->
	  let (dx1,dx2) = match dxlist with
	  | (dx1,dx2) :: _ -> (dx1*rad*2, dx2*rad*2)
	  | _ -> (0, 0)
	  in
	  set_color black;
	  if g <> Vide then draw_segments [| (x,y,x+dx1,y-dy) |];
	  if d <> Vide then draw_segments [| (x,y,x+dx2,y-dy) |];
	  begin try
	    let clef = Arbres.clef v in
	    let (fcolor, bcolor) = Arbres.dcolor couleurs v in
	    set_color bcolor;
	    fill_circle x y rad;
	    set_color black;
	    draw_circle x y rad;
	    set_color fcolor;
	    let str = Printf.sprintf "%d" clef in
	    let (w,h) = text_size str in
	    moveto (x-w/2) (y-h/2);
	    draw_string str;
	  with
	    Not_found -> ()
	  end;
	  aux (x+dx1) (y-dy) g;
	  aux (x+dx2) (y-dy) d
    in aux x y (calculer_dx arbre)

  let dessiner_std arbre =
    open_graph "";
    set_text_size 16;
    let (_, hauteur) = calculer_dim arbre in
    dessiner 60 16 ((size_x ()+1)/2) ((size_y ()+60*hauteur+1)/2) arbre
    
end

(****************************************************************************************************)

module Naif = MakeDessineur(Arbres_naifs)
module AVL = MakeDessineur(Arbres_avl)
module RN = MakeDessineur(Arbres_rn)

(****************************************************************************************************)
